#!/usr/local/bin/perl
# ESA (C) 2000-2018 
#
# This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
#
#    SAS is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    SAS is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
# 
# Usage :
#
# ds9tomask 
#        eventsets="EVENT LISTs filenames of the EPIC imaging observation"
#        imagesets="EPIC Images in band 8"
#        regionfile = input DS9 region file
#
# ------------------------------------------------------
# Define variables
# ------------------------------------------------------ 

use SAS;
use strict;

sub ds9tomask{

    my($att,$attCol);
    my($c1,$c2,$command,$colnum,$colNumEVT,$colNumPSD);
    my($evPseudo,$evselBinSizePar,$exist,$excl_filter, );
    my($filterXcol, $filterYcol,$final_filter, $final_mask);
    my($i,$ia,$ifilter,$img,$imgCkeys,$incl_filter, $inPixSize,$inst);
    my($key,$l1,$l2,$l1_hf,$l2_hf,$na,$nEvtSets);
    my($radius,$regfile,$saserr,$sasmess,$srcpixval,$tcdltX);
    my($xColPixSize);
    
    
    my @evtlists   = ();
    my @imgs       = ();
    my @tmpmasks   = ();
    my @tmpmasks2  = ();
    my @filters    = ();
    my @exclfilters = ();
    my @manfilters = ();
    my @sinfilters = ();
    my @srcfilters = ();
    my @pseudoAtt = ("FILTER","INSTRUME","TELESCOP","AVRG_PNT","RADECSYS","REFYCUNI","REFXCUNI","REFYCTYP");
    push(@pseudoAtt,"PA_PNT","RA_PNT","DEC_PNT","EQUINOX","RA_NOM","DEC_NOM","RA_OBJ","DEC_OBJ");
    push(@pseudoAtt,"REFYCUNI","REFYDMAX","REFYDMIN","REFYLMAX","REFYLMIN");
    push(@pseudoAtt,"REFYCDLT","REFYCRVL","REFYCRPX","REFYCTYP","REFXCUNI");
    push(@pseudoAtt,"REFXDMAX","REFXDMIN","REFXLMAX","REFXLMIN");
    my @pseudoAttCol = ("TCUNI","TCTYP","TCDLT","TCRVL","TCRPX","TDMAX","TDMIN","TLMAX","TLMIN");
    my @colnames      = ("DETX","DETY","X","Y");
    
    my %colNums = ();
    
# ------------------------------------------------------
# Read input parameters
# ------------------------------------------------------
    
    $regfile = stringParameter( "regionfile" );
    $srcpixval = intParameter( "srcpixvalue" );
    
    $nEvtSets = parameterCount( "eventsets" );
    for( $i=0; $i<$nEvtSets; $i++ ) {
	$evtlists[$i] = stringParameter( "eventsets", $i );
	$imgs[$i]     = stringParameter( "imagesets",$i );
	$tmpmasks[$i] = stringParameter( "tmpmasks",$i );
	$tmpmasks2[$i] = $tmpmasks[$i].2;
    }
    my $tmpEvtPseudo = stringParameter( "tmpevt" );
    my $maskset = stringParameter( "maskset" );
    
    my @tempfiles=( $tmpEvtPseudo );
    push( @tempfiles,@tmpmasks, @tmpmasks );
    
#--------------------------------------------------------
# Find out TCRPX, TCDLT attributes for column X
#--------------------------------------------------------
    
    $sasmess = " ";
    SAS::message($SAS::AppMsg,$SAS::SparseMsg,$sasmess);

    for ( $i=0; $i<$nEvtSets; $i++ ){
	
	$sasmess = "Processing evt: $evtlists[$i]\n";
	SAS::message($SAS::AppMsg,$SAS::SparseMsg,$sasmess);
	
	# first get column number
	if ( system( "fcolpar infile=$evtlists[$i]\[EVENTS\] colname='X'" ) ) {
	    clearTemporary( @tempfiles );
	    $saserr = "Cannot establish column number for X in $evtlists[$i]";
	    SAS::error( "ErrFcolpar",$saserr );
	}
	$colnum = `pget fcolpar colnum`;
	chomp( $colnum );
	
	$att = "TCDLT" . $colnum;
	if ( system( "fkeypar fits=$evtlists[$i]\[EVENTS\] key=$att" ) ) {
	    clearTemporary( @tempfiles );
	    $saserr = "Cannot run fkeypar to read $att from $evtlists[$i]";
	    SAS::error( "ErrFkeypar",$saserr );
	}
	
	$tcdltX = `pget fkeypar value`;
	chomp( $tcdltX );
	$xColPixSize = abs( $tcdltX );
	
	
	# ------------------------------------------------------
	# Obtain some pixel size/position WCS attributes from imageset
	# ------------------------------------------------------
	
	$imgCkeys = $imgs[$i] . "+0";
	if ( system( "fkeypar fits=$imgCkeys key='CDELT1'" ) ) {
	    clearTemporary( @tempfiles );
	    $saserr = "Cannot run fkeypar to read CDELT1 from $imgs[$i]";
	    SAS::error("ErrFkeypar",$saserr);
	}
	$exist = `pget fkeypar exist`;
	unless( $exist =~ /yes/ ) {
	    $imgCkeys = $imgs[$i]."+1";
	    if( system( "fkeypar fits=$imgCkeys key='CDELT1'" ) ) {
		clearTemporary( @tempfiles );
		$saserr = "Cannot run fkeypar to read CDELT1 from $imgs[$i]";
		SAS::error("ErrFkeypar",$saserr);
	    }
	}
	$inPixSize = `pget fkeypar value`;
	chomp( $inPixSize );
	$inPixSize = abs( $inPixSize );
	
	
	# binSize parameter for evselect when making images
	$evselBinSizePar = $inPixSize / $xColPixSize;
	$evselBinSizePar = sprintf( "%.0f", $evselBinSizePar );
	
	
	# --------------------------------------------------------
	#  2) Create a mask to "mark" ds9 regions in input image
	#---------------------------------------------------------
	#
	#     2.1) Create pseudo evt list with X/Y/DETX/DETY columns
	#          (run makepseudoevt F90 module) --> $tmpEvtPseudo
	#
	
	if ( system( "makepseudoevt set=$tmpEvtPseudo imageset=$imgs[$i] inset=$evtlists[$i]")){
	    clearTemporary( @tempfiles );
	    SAS::error("ErrPseudo","Cannot create Pseudo EVT");
	}
	
	#
	#     2.2) Write WCS keywords (from input EVT list) to PSEUDO event list
	#
	#          a) Add  Attributes to Pseudo EVT list
	
	$sasmess="Adding Single Atrributes to PSEUDO EVT";
	SAS::message($SAS::AppMsg,$SAS::SparseMsg,$sasmess);
	$ia = 0;
	$na = @pseudoAtt;
	foreach my $att ( @pseudoAtt ){
	    $ia++;
	    $sasmess="   .... $ia of $na ($att)";
	    SAS::message($SAS::AppMsg,$SAS::SparseMsg,$sasmess);
	    # get attribute value from initial EVT file
	    if( system( "fkeypar fits=$evtlists[$i]\[EVENTS\] key=$att" ) ) {
		clearTemporary( @tempfiles );
		$saserr = "Cannot run fkeypar to read $att from eventlist";
		SAS::error("ErrFkeypar",$saserr);
	    }
	    $exist = `pget fkeypar exist`;
	    ($exist =~ /no/) && (next);
	    $key = `pget fkeypar value`;
	    chomp( $key );
	    ( $att =~ /INSTRUME/ ) && ( $inst = $key );
	    # add attribute value to PSEUDO EVT file
	    if( system( "fparkey fits=$tmpEvtPseudo\[EVENTS\] add=yes value=$key key=$att" ) ) {
		clearTemporary( @tempfiles );
		$saserr = "Cannot run fparkey to add $att to Pseudo event list";
		SAS::error("ErrFparkey",$saserr);
	    }
	}
      
	#          b) Add Column attributes to PSEUDO EVT
	$sasmess="Getting Column Numbers in EVT and PSEUDO EVT";
	SAS::message($SAS::AppMsg,$SAS::SparseMsg,$sasmess);
	
	foreach my $colname ( @colnames ){
	    if ( system( "fcolpar infile=$evtlists[$i]\[EVENTS\] colname='$colname'" ) ) {
		clearTemporary( @tempfiles );
		$saserr = "Cannot establish column number for $colname in $evtlists[$i]";
		SAS::error( "ErrFcolpar",$saserr );
	    }
	    $colNumEVT = `pget fcolpar colnum`;
	    chomp( $colNumEVT );
	    
	    if ( system( "fcolpar infile=$tmpEvtPseudo\[EVENTS\] colname='$colname'" ) ) {
		clearTemporary( @tempfiles );
		$saserr = "Cannot establish column number for $colname in PSEUDO EVT $tmpEvtPseudo";
		SAS::error( "ErrFcolpar",$saserr );
	    }
	    $colNumPSD = `pget fcolpar colnum`;
	    chomp( $colNumPSD );
	    
	    $colNums{$colname} = [$colNumEVT, $colNumPSD];
	}
	
	
	$sasmess="Adding Column Atrributes to PSEUDO EVT";
	SAS::message($SAS::AppMsg,$SAS::SparseMsg,$sasmess);
	$ia = 0;
	$na = @pseudoAttCol;
	foreach my $att ( @pseudoAttCol ){
	    $ia++;
	    $sasmess="   .... $ia of $na ($att)";
	    SAS::message($SAS::AppMsg,$SAS::SparseMsg,$sasmess);
	    foreach my $colname ( @colnames ){
		$colNumEVT = $colNums{$colname}[0];
		$colNumPSD = $colNums{$colname}[1];
		
		# get column attribute name and value in initial EVT file
		$attCol = $att . "$colNumEVT";
		if( system( "fkeypar fits=$evtlists[$i]\[EVENTS\] key=$attCol" ) ) {
		    clearTemporary( @tempfiles );
		    $saserr = "Cannot run fkeypar to read $attCol from eventlist";
		    SAS::error("ErrFkeypar",$saserr);
		}
		$exist = `pget fkeypar exist`;
		($exist =~ /yes/) || (next);
		$key = `pget fkeypar value`;
		chomp( $key );
		#$sasmess="   ............... reading $attCol";
		#SAS::message($SAS::AppMsg,$SAS::SparseMsg,$sasmess);
		
		# add column attribute value to PSEUDO EVT file
		$attCol = $att . "$colNumPSD";
		if( system( "fparkey add=yes fits=$tmpEvtPseudo\[EVENTS\] value=$key key=$attCol" ) ) {
		    clearTemporary( @tempfiles );
		    $saserr = "Cannot run fparkey to add $attCol to Pseudo event list";
		    SAS::error("ErrFparkey",$saserr);
		}
		#$sasmess="   ............... writting $attCol";
		#SAS::message($SAS::AppMsg,$SAS::SparseMsg,$sasmess);
	    }
	}
	
	#
	#          c) Add XMM attributes so that evselect does not complain
	#
	
	$sasmess="Adding XMM Atrributes to PSEUDO EVT";
	SAS::message($SAS::AppMsg,$SAS::SparseMsg,$sasmess);
	if( system( "fparkey add=yes value='XMM' fits=$tmpEvtPseudo+0 key=TELESCOP" )){
	    clearTemporary( @tempfiles );
	    $saserr = "Cannot run fparkey to add TELESCOP to Pseudo event list\n";
	    SAS::error("ErrFparkey",$saserr);
	}
	if( system( "fparkey add=yes value='XMM' fits=$tmpEvtPseudo+1 key=TELESCOP" )){
	    clearTemporary( @tempfiles );
	    $saserr = "Cannot run fparkey to add TELESCOP to Pseudo event list\n";
	    SAS::error("ErrFparkey",$saserr);
	}
	if( system("fparkey add=yes value=$inst fits=$tmpEvtPseudo+0 key=INSTRUME" )){
	    clearTemporary( @tempfiles );
	    $saserr = "Cannot run fparkey to add INSTRUME to Pseudo event list\n";
	    SAS::error("ErrFparkey",$saserr);
	}
	if( system( "fparkey add=yes value=$inst fits=$tmpEvtPseudo+1 key=INSTRUME" )){
	    clearTemporary( @tempfiles );
	    $saserr = "Cannot run fparkey to add INSTRUME to Pseudo event list\n";
	    SAS::error("ErrFparkey",$saserr);
	}
	
	#
	#     2.3) Filter pseudo evt list with ds9 region and save
	#         image (=mask) ---> ds9mask (same size/dimensions as input image)
	#         Mask: "1" inside region and "2" otherwise
	
	# set coordinate type in DS9 region file
	$filterXcol = "X";
	$filterYcol = "Y";
	if(`grep detector $regfile|wc -l`){
	    $filterXcol = "DETX";
	    $filterYcol = "DETY";      
	}
	
	$sasmess="Defining Filters from input region file";
	SAS::message($SAS::AppMsg,$SAS::SparseMsg,$sasmess);
	
	#
	#  read DS9 regions file
	#
	@filters    = ();
	@manfilters = ();
	@sinfilters = ();
	@srcfilters = ();
	
	open(REG,"$regfile");
	@filters = <REG>;
	close(REG);
	
	
	#use "manually" defined regions as they are (exclude)
	@manfilters = grep(/tag=\{man\}/,@filters);
	#use "single" defined regions transf. them into circles (R=10") to exclude
	@sinfilters = grep(/tag=\{sin\}/,@filters);
	#use "src" defined regions transf. them into circles (R=10") to INCLUDE
	if ( $srcpixval == 2 ){
	    @srcfilters = grep(/tag=\{src\}/,@filters);
	} #otherwise (=0), do nothing
	
	$ifilter = 0;
	
	$final_filter = "";
	foreach my $filter ( @manfilters ){
	    chomp($filter);
	    # DS9   : box(xCenter,yCenter,Width, height, angle)
	    # DSSLIB: box(xCenter,yCenter,xHalfWidth,yHalfWidth,rotation,Xcolumn,Ycolumn)
	    if( $filter =~ /box/ ) {
		( $c1, $c2, $l1, $l2 ) = split( /,/, $filter );
		$l1_hf = $l1/2.;
		$l2_hf = $l2/2.;
		$filter =~ s/$l1/$l1_hf/;
		$filter =~ s/$l2/$l2_hf/;
	    }
	    $filter =~ s/\).*/,$filterXcol,$filterYcol\)/;
	    $filter = "\(!\($filter\)\)";
	}
	
	foreach my $filter ( @sinfilters ){
	    chomp($filter);
	    $radius = 10.0/0.05; # both X,Y and DETX,DETY are in units of 0.05 arcs
	    #             box( nnnn.nnn , nnnn.nnn    )
	    $filter =~ s/^.*\((\d+\.*\d*),(\d+\.*\d*).*\).*/circle\($1,$2,$radius,$filterXcol,$filterYcol\)/;
	    $filter = "\(!\($filter\)\)";
	}
	
	foreach my $filter ( @srcfilters ){
	    chomp($filter);
	    $radius = 10.0/0.05; # both X,Y and DETX,DETY are in units of 0.05 arcs
	    #           circle( nnnn.nnn , nnnn.nnn   )
	    $filter =~ s/^.*\((\d+\.*\d*),(\d+\.*\d*).*\).*/circle\($1,$2,$radius,$filterXcol,$filterYcol\)/;
	    $filter = "\($filter\)";
	}
	
	@exclfilters=();
	@exclfilters = ( @manfilters, @sinfilters );
	
	$excl_filter = join(" \&\& ",@exclfilters);
	$incl_filter = join(" \|\| ",@srcfilters);
	$final_filter = $excl_filter;
	( @srcfilters ) && ($final_filter .= " || ".$incl_filter );

	#$sasmess="Filtering temporary event file with filter =$final_filter";
	$sasmess="Filtering PSEUDO EVT file ";
	SAS::message($SAS::AppMsg,$SAS::SparseMsg,$sasmess);
	
	# if srcpixval=2 : create two masks
	#     ** one with "1" only in the 'src' region
	#     ** one with the full filter
	# and then add them, so that pixels=2 in the 'src' region
	
	$sasmess="Creating mask with evselect from PSEUDO EVT";
	SAS::message($SAS::AppMsg,$SAS::SparseMsg,$sasmess);
	if( system( "evselect table=$tmpEvtPseudo:EVENTS  expression=\"$final_filter\" withfilteredset=false keepfilteroutput=false destruct=false writedss=false updateexposure=false withimageset=true imageset=$tmpmasks[$i] xcolumn='X' ycolumn='Y' imagebinning=binSize ximagebinsize=$evselBinSizePar yimagebinsize=$evselBinSizePar" ) ) {
	    clearTemporary( @tempfiles );
	    $saserr = "Cannot run EVSELECT to filter input evt list with region filter\n";
	    SAS::error("ErrEvselect",$saserr);
	}
	
	if ( $srcpixval == 2 && @srcfilters){
	    $sasmess="Creating Second Mask for Included Manual (SRC)regions";
	    SAS::message($SAS::AppMsg,$SAS::SparseMsg,$sasmess);
	    
	    if( system( "evselect table=$tmpEvtPseudo:EVENTS  expression=\"$incl_filter\" withfilteredset=false keepfilteroutput=false destruct=false writedss=false updateexposure=false withimageset=true imageset=$tmpmasks[$i].2 xcolumn='X' ycolumn='Y' imagebinning=binSize ximagebinsize=$evselBinSizePar yimagebinsize=$evselBinSizePar" ) ) {
		clearTemporary( @tempfiles );
		$saserr = "Cannot run EVSELECT to filter input evt list with region filter\n";
		SAS::error("ErrEvselect",$saserr);
	    } # src filter mask
	    
	    if( system( "farith infil1=$tmpmasks[$i] infil2=$tmpmasks[$i].2 outfil=\!$tmpmasks[$i] ops=ADD" ) ) {
		clearTemporary( @tempfiles );
		$saserr = "Cannot run FARITH to add 'src' masks\n";
		SAS::error("ErrFarith",$saserr);
	    }
	} #srcpixval=2
    } # end list of evt files
    
    my $masks_string = join( " ",@tmpmasks );
    if (@tmpmasks == 1){

	$maskset=$tmpmasks[0];

    }else{

	$sasmess="Doing 'emosaic' over list of output masks";
	SAS::message($SAS::AppMsg,$SAS::SparseMsg,$sasmess);

	if( system( "emosaic imagesets=\"$masks_string\" withexposure=no mosaicedset=$maskset sampling=point forceuniformkwds=yes" ) ) {
	    clearTemporary( @tempfiles );
	    $saserr = "Cannot run EMOSAIC to add instrument masks:$!\n";
	    SAS::error("ErrEmosaic",$saserr);
	}
	$sasmess="Averaging masks";
	SAS::message($SAS::AppMsg,$SAS::SparseMsg,$sasmess);
	if( system( "fcarith infile=$maskset const=$nEvtSets outfil=\!$maskset ops=DIV" ) ) {
	    clearTemporary( @tempfiles );
	    $saserr = "Cannot run FCARITH to divide by the number of masks:$!\n";
	    SAS::error("ErrFcarith",$saserr);
	}
    }
}


sub clearTemporary{
    foreach (@_){
	(-e $_) && unlink($_);
    }
}
