! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!+
! Module: makepseudoevt_mod
!
! Authors:  M.T. Ceballos -  SSC (IFCA) 
!           F.J. Carrera -  SSC (IFCA)  
!           X. Barcons - SSC (IFCA)
!          
! Description: Creates the pseudoEvt file with X/Y, DETX/DETY columns
!              from input image
!              
!

module makepseudoevt_mod

  use dal
  use cal
  use coordinate  ! from ssclib for coordinate conversion	
  use wcs_aux
  use constants   !DegToRad=pi/180
  use param  
  use errorHandling
  
  implicit none 
  
  private 

  public makepseudoevt

contains 

  !---------------------------------------------------------------------
  subroutine makepseudoevt
  !---------------------------------------------------------------------
    implicit none

    real(kind=single),dimension(:),allocatable :: xPixelsImg,yPixelsImg
    real(kind=single),dimension(:),allocatable :: xPixelsCol,yPixelsCol
    real(kind=double),dimension(:),allocatable :: detxCol,detyCol
    real(kind=double),dimension(:),allocatable :: tanX,tanY
    real(kind=double), dimension(:), pointer :: coX_Data,coY_Data,coDETX_Data,coDETY_Data
    real(kind=double) :: expStartTimeStamp

    integer(kind=Int32) :: i,j,ipix,irow,nrows,instrumentID
    integer(kind=Int32),dimension(:), pointer :: shape_inImage

    character(len=256) :: outEvt,inImage,inEvt,dateTimeStr
    
    type(DataSetT) :: set_outEvt,set_inEvt,set_inImage
    type(TableT):: tab_outEvt,tab_inEvt
    type(ArrayT) :: arr_inImage
    type(ColumnT) :: coX_outEvt,coY_outEvt, coDETX_outEvt,coDETY_outEvt
    type(SpacecraftAttitudeType) :: scAttitude
    type(WcsT):: wcsImg,wcsColXY

!   read input parameters from param file
    
    outEvt=stringParameter("set")
    inImage=stringParameter("imageset")
    inEvt=stringParameter("inset")
    
    set_inImage=dataSet(inImage,Read)
    arr_inImage=array(set_inImage,"PRIMARY",Read)
    shape_inImage => dimensions(arr_inImage)
    set_inEvt=dataSet(inEvt,read)
    tab_inEvt=table(set_inEvt,"EVENTS",Read)
    nrows=shape_inImage(0)*shape_inImage(1)

!   Create columns in output event list: X,Y,DETX,DETY

    set_outEvt=dataSet(outEvt,CREATE)
    tab_outEvt=addtable(set_outEvt,"EVENTS",nrows)
    coX_outEvt=addcolumn(tab_outEvt,"X",Real64) 
    coY_outEvt=addcolumn(tab_outEvt,"Y",Real64)
    coDETX_outEvt=addcolumn(tab_outEvt,"DETX",Real64) 
    coDETY_outEvt=addcolumn(tab_outEvt,"DETY",Real64)
    coX_Data=>Real64Data(coX_outEvt)
    coY_Data=>Real64Data(coY_outEvt)
    coDETX_Data=>Real64Data(coDETX_outEvt)
    coDETY_Data=>Real64Data(coDETY_outEvt)

!
!   Reading WCS info in IMAGE and EVTLIST for coordinate transformation
!
    call readWcs(set_inImage, arr_inImage, wcsImg)
    call readWcs(tab_inEvt,"X","Y", wcsColXY)


    if(stringAttribute(set_inImage,"INSTRUME") == "EMOS1")then
       instrumentID=4
    elseif(stringAttribute(set_inImage,"INSTRUME") == "EMOS2")then
       instrumentID=5
    elseif(stringAttribute(set_inImage,"INSTRUME") == "EPN")then
       instrumentID=6
    endif

    call CAL_setState(instrument=instrumentID)

    dateTimeStr=stringAttribute(set_inImage,"DATE-OBS")
    scAttitude%ra=real32Attribute(set_inImage,"RA_PNT")
    scAttitude%dec=real32Attribute(set_inImage,"DEC_PNT")
    scAttitude%apos=real32Attribute(set_inImage,"PA_PNT")

    expStartTimeStamp=toTimeTag(dateTimeStr)
    scAttitude%ra=scAttitude%ra*DegToRad
    scAttitude%dec=scAttitude%dec*DegToRad
    scAttitude%apos=scAttitude%apos*DegToRad

    call release(set_inImage)


!   Allocate arrays to store required coordinates

    allocate(xPixelsImg(0:nrows-1))
    allocate(yPixelsImg(0:nrows-1))
    allocate(xPixelsCol(0:nrows-1))
    allocate(yPixelsCol(0:nrows-1))
    allocate(detxCol(0:nrows-1))
    allocate(detyCol(0:nrows-1))
    allocate(tanX(0:nrows-1))
    allocate(tanY(0:nrows-1))

!   Populate initial arrays indexing image pixels 
    ipix=0
    do i=1,shape_inImage(0)
       do j=1,shape_inImage(1)
          xPixelsImg(ipix)=i
          yPixelsImg(ipix)=j
          ipix=ipix+1
       enddo
    enddo

!
!   Coordinate Transformation
!   -------------------------


!   Transforms pixels coordinates (i,j) in the image to tangent plane 
!   coordinates in rad using CRPIX1,CRPIX2,CDELT1,CDELT2 and CROTA2
    call pixelsToTan(xPixelsImg, yPixelsImg, wcsImg, tanX, tanY)
    deallocate(xPixelsImg,yPixelsImg)


!   From image tangent coordinates to X,Y evtlist columns coordinates using
!   TCRPIX"X",TCRPIX"Y",TCDLT"X",TCDLT"Y" and CROTA2
    call tanToPixels(wcsColXY,tanX,tanY,xPixelsCol,yPixelsCol)

!   From image tangent coordinates to DETX,DETY evtlist columns coordinates 
    call tanToDet(tanX,tanY,wcsImg,scAttitude,expStartTimeStamp,detxCol,detyCol)

    deallocate(tanX, tanY)


!   Fill in columns
    do irow=0,nrows-1
       coX_Data(irow)=xPixelsCol(irow)
       coY_Data(irow)=yPixelsCol(irow)
       coDETX_Data(irow)=detxCol(irow)
       coDETY_Data(irow)=detyCol(irow)
    enddo
    deallocate(xPixelsCol,yPixelsCol,detxCol,detyCol)


    call release(set_outEvt);

  end subroutine makepseudoevt

end module makepseudoevt_mod
